@@ -3,6 +3,7 @@ import gc
 import inspect
 import itertools
 import functools
+import re
 import sys
 
 from timeit import default_timer
@@ -109,6 +110,7 @@ def print_performance(f=None, repeat=5):
                 tic  = default_timer()
                 retv = _f_(*args, **kwargs)
                 toc  = default_timer()
+
                 elapsed = -(tic - toc)
 
                 if best is None:
@@ -180,20 +182,20 @@ def styled(message,
             if not (e.startswith('\x1b') and e.endswith('m')):
                 raise KeyError('styled: invalid effect: {}'.format(e))
 
-    if not is_utf_console:       return message
-    if not color and not effect: return message
+    if not is_utf_console:
+        return message
+
+    if not color and not effect:
+        return message
 
     ansi_color  = __color_codes__.get(color, color)
     ansi_effect = ''.join(__effect_codes__.get(e, e) for e in effect)
-
     __effect_start__  = ansi_color + ansi_effect
 
     # only apply new style to any unstyled parts of message
     _message_ = message.replace(__effect_end__,
                                 __effect_end__ + __effect_start__)
-    _message_ = (__effect_start__ +
-                 _message_ +
-                 __effect_end__)
+    _message_ = '{}{}{}'.format(__effect_start__, _message_, __effect_end__)
 
     return _message_
 
@@ -371,11 +373,9 @@ def format_seconds(s):
 
 def function_parameters(f):
     """
-
     def function(something, *, also, also_1=None):
         pass
 
-
     Arguments(args=['something', 'also', 'also_1'],
               varargs=None,
               varkw=None) = inspect.getargs(function.__code__)
@@ -414,6 +414,32 @@ def function_parameters(f):
     return n_params
 
 
+# def function_name(f):
+#     try:
+#         fname = f.__qualname__
+#     except AttributeError:
+#         try:
+#             if isinstance(f, property):
+#                 fname = str(f.fget).split(' ')[1]
+#             else:
+#                 fname = str(f)
+#         except:
+#             fname = str(f)
+#
+#     if '.' in fname:         # probably a class method
+#         return fname
+#
+#     try:
+#         modulename = f.__module__
+#     except AttributeError:
+#         modulename = str(f)
+#
+#     if '.' in fname:
+#         modulename = modulename.split('.')[-1]
+#
+#     return '{}.{}'.format(modulename, fname)
+
+
 def function_name(f):
     try:
         name = f.__qualname__
@@ -435,6 +461,7 @@ def function_name(f):
     return '{}.{}'.format(modulename, name)
 
 
+
 def object_name(o):
     try:                   return o.__name__
     except AttributeError: pass
@@ -446,6 +473,25 @@ def object_name(o):
     except AttributeError: return ''
 
 
+# noinspection DuplicatedCode
+def snake_case(s):
+    """ eg:
+        'some_value' = snake_case('someValue')
+    """
+    camel_re = re.compile('''
+        (?<=[a-z])[A-Z](?=[a-z])
+    ''', re.VERBOSE)
+
+    if s[0] == s[0].upper():
+        s = ''.join((s[0].lower(), s[1:]))
+
+    for i, match in enumerate(camel_re.finditer(s)):
+        c = match.span()[0] + i
+        p = '_' + s[c].lower()
+        s = s[:c] + p + s[c + 1:]
+
+    return s.lower().strip()
+
 
 
 
